@@ -323,7 +323,7 @@ SensAnalysisMLP <- function(MLP.fit, .returnSens = TRUE, plot = TRUE, .rawSens =
 #'
 #' @method SensAnalysisMLP default
 SensAnalysisMLP.default <- function(MLP.fit, .returnSens = TRUE, plot = TRUE, .rawSens = FALSE, trData,
-                                    actfunc = c('linear', 'sigmoid','linear'),preProc = NULL,
+                                    actfunc = NULL,preProc = NULL,
                                     terms = NULL, ...) {
   ### Things needed for calculating the sensibilities:
   #   - Structure of the model  -> MLP.fit$n
@@ -393,43 +393,34 @@ SensAnalysisMLP.default <- function(MLP.fit, .returnSens = TRUE, plot = TRUE, .r
   ActivationFunction <- lapply(actfunc, NeuralSens::ActFunc)
   DerActivationFunction <- lapply(actfunc, NeuralSens::DerActFunc)
 
-  # Build the jacobian of the first layer
-  D[[1]] <- array(diag(ncol(TestData)),c(ncol(TestData),ncol(TestData),nrow(TestData)))
   W[[1]] <- array(c(0,rep(1,ncol(TestData))),c(ncol(TestData)+1,1,nrow(TestData)))
   # For each row in the TestData
   Z[[1]] <- as.matrix(TestData)
-  O[[1]] <- apply(Z[[1]],c(1,2),
-                  function(x){unlist(lapply(x, ActivationFunction[[1]]))})
-
+  O[[1]] <- ActivationFunction[[1]](Z[[1]])
+  # Build the jacobian of the first layer
+  D[[1]] <- array(NA, dim=c(mlpstr[1], mlpstr[1], nrow(TestData)))
+  for(irow in 1:nrow(TestData)){
+    D[[1]][,,irow] <- DerActivationFunction[[1]](Z[[1]][irow,])
+  }
   # For each layer, calculate the input to the activation functions of each layer
   # This inputs are gonna be used to calculate the derivatives and the output of each layer
   for (l in 2:length(mlpstr)){
     W[[l]] <- data.matrix(as.data.frame(wts[(sum(mlpstr[1:(l-1)])-mlpstr[1]+1):(sum(mlpstr[1:l])-mlpstr[1])]))
     Z[[l]] <- cbind(1, O[[l-1]]) %*% W[[l]]
-    O[[l]] <- apply(Z[[l]],c(1,2),
-                    function(x){unlist(lapply(x, ActivationFunction[[l]]))})
+    O[[l]] <- ActivationFunction[[l]](Z[[l]])
 
     # Detect if it's a vector because we need it in row vectors and in r a vector is a column
     m <- W[[l]][2:nrow(W[[l]]),]
-    z <- apply(Z[[l-1]],c(1,2),
-               function(x){unlist(lapply(x, DerActivationFunction[[l-1]]))})
-    # d__ <- do.call(rbind,
-    #                lapply(seq_len(dim(D[[l-1]])[3]),
-    #                       function(i) t(m*z[i,])%*%D[[l-1]][,,i]))
-    d_ <- array(NA, dim=c(mlpstr[l], mlpstr[1], nrow(TestData)))
+    D[[l]]<- array(NA, dim=c(mlpstr[1], mlpstr[l], nrow(TestData)))
+    # Chain rule for calculate the derivatives between layers
     for(irow in 1:nrow(TestData)){
-      d_[,,irow] <- t(m*z[irow,]) %*% D[[l-1]][,,irow]
+      z <- DerActivationFunction[[l]](Z[[l]][irow,])
+      D[[l]][,,irow] <- D[[l-1]][,,irow] %*% m %*% z
     }
-    D[[l]] <- d_
   }
   # Output of the neural network is the output of the last layer
   out <- O[[length(O)]]
-  # Apply last derivative
-  z <- apply(Z[[l]],c(1,2),
-             function(x){unlist(lapply(x, DerActivationFunction[[l]]))})
-  a <- aperm(D[[length(D)]],c(3,2,1))
-  der <- array(do.call(rbind,lapply(seq_len(dim(a)[1]),function(i) a[i,,]*z[i,])),
-               c(dim(a)))
+  der <- aperm(D[[l]],c(3,1,2))
   sens <-
     data.frame(
       varNames = varnames,
@@ -614,7 +605,7 @@ SensAnalysisMLP.H2OMultinomialModel <- function(MLP.fit, .returnSens = TRUE, plo
   # Change the name of the output in trData
   if(MLP.fit@parameters$y %in% names(trData)) {
     names(trData)[which(MLP.fit@parameters$y == names(trData))] <- ".outcome"
-  } else {
+  } else if (!".outcome" %in% names(trData)) {
     stop(paste0("Output ",MLP.fit@parameters$y," has not been found in training data"))
   }
   trData <- trData[,c(".outcome",MLP.fit@parameters$x)]
@@ -656,7 +647,7 @@ SensAnalysisMLP.H2OMultinomialModel <- function(MLP.fit, .returnSens = TRUE, plo
              stop("SensAnalysisMLP function is not ready for Maxout layers")
            },
            Softmax = {
-             return("sigmoid")
+             return("softmax")
            },
            {
              stop("SensAnalysisMLP is not ready for the activation function used")
@@ -761,7 +752,7 @@ SensAnalysisMLP.H2ORegressionModel <- function(MLP.fit, .returnSens = TRUE, plot
   # Change the name of the output in trData
   if(MLP.fit@parameters$y %in% names(trData)) {
     names(trData)[which(MLP.fit@parameters$y == names(trData))] <- ".outcome"
-  } else {
+  } else if (!".outcome" %in% names(trData)) {
     stop(paste0("Output ",MLP.fit@parameters$y," has not been found in training data"))
   }
   trData <- trData[,c(".outcome",MLP.fit@parameters$x)]
@@ -976,3 +967,158 @@ SensAnalysisMLP.nnet <- function(MLP.fit, .returnSens = TRUE, plot = TRUE, .rawS
                           terms = terms,
                           plot = plot, ...)
 }
+
+#' @rdname SensAnalysisMLP
+#'
+#' @export
+#'
+#' @method SensAnalysisMLP nnetar
+SensAnalysisMLP.nnetar <- function(MLP.fit, .returnSens = TRUE, plot = TRUE, .rawSens = FALSE,...) {
+  # Create the lags in the trData
+  args = list(...)
+  if (!is.null(MLP.fit$xreg)) {
+    if ("trData" %in% names(args)) {
+      xreg <- args$trData[,attr(MLP.fit$xreg,"dimnames")[[2]]]
+      if(!".outcome" %in% names(args$trData)) {
+        stop("Change the name of the output variable to '.outcome'")
+      }
+      outcome <- args$trData$.outcome
+    } else {
+      xreg <- as.data.frame(MLP.fit$xreg)[MLP.fit$subset,]
+      outcome <- MLP.fit$x[MLP.fit$subset]
+    }
+  } else {
+    outcome <- MLP.fit$x[MLP.fit$subset]
+  }
+  # Scale the regressors
+  if (!is.null(MLP.fit$scalexreg)) {
+    for (i in 1:ncol(xreg)) {
+      varname <- names(xreg)[[i]]
+      indexscale <- which(attr(MLP.fit$scalexreg$center,"names") == varname)
+      xreg[[i]] <- (xreg[[i]] - MLP.fit$scalexreg$center[indexscale])/MLP.fit$scalexreg$scale[indexscale]
+    }
+  }
+  # Scale the output
+  if (!is.null(MLP.fit$scalex)) {
+    outcome <- (outcome - MLP.fit$scalex$center)/MLP.fit$scalex$scale
+  }
+
+  # Create lagged outcome as input
+  ylagged <- NULL
+  p <- MLP.fit$p
+  P <- MLP.fit$P
+  m <- MLP.fit$m
+  if (P > 0) {
+    lags <- sort(unique(c(1:p, m * (1:P))))
+  } else {
+    lags <- 1:p
+  }
+  for (i in lags) {
+    ylagged[[i]] <- Hmisc::Lag(outcome, i)
+    names(ylagged)[i] <- paste0(".outcome_Lag",as.character(i))
+  }
+  ylagged <- as.data.frame(ylagged[lags])
+
+  if (!is.null(MLP.fit$xreg)) {
+    trData <- cbind(ylagged, xreg, as.data.frame(outcome))
+    varnames <- c(names(trData)[1:MLP.fit$p],names(as.data.frame(MLP.fit$xreg)))
+  } else {
+    trData <- cbind(ylagged, as.data.frame(outcome))
+    varnames <- names(trData)[1:MLP.fit$p]
+  }
+
+  names(trData)[ncol(trData)] <- ".outcome"
+  # Get rid of rows with NAs
+  trData <- trData[stats::complete.cases(trData),]
+
+  # For a nnet nnet
+  finalModel <- NULL
+  sensitivities <- list()
+  finalModel$n <- MLP.fit$model[[1]]$n
+  actfun <- c("linear","sigmoid",
+              ifelse(is.factor(trData$.outcome),"sigmoid","linear"))
+  finalModel$coefnames <- varnames
+  # Apply default function to all the models in the nnetar object
+  for (i in 1:length(MLP.fit$model)) {
+    finalModel$wts <- MLP.fit$model[[1]]$wts
+    sensitivities[[i]] <-  SensAnalysisMLP.default(finalModel,
+                                              trData = trData,
+                                              actfunc = actfun,
+                                              .returnSens = TRUE,
+                                              .rawSens = TRUE,
+                                              preProc = NULL,
+                                              terms = NULL,
+                                              plot = FALSE)
+  }
+
+  sensitivities <- as.data.frame(do.call("rbind",lapply(sensitivities,
+                                               function(x) {
+                                                 as.data.frame(x[1:dim(x)[1],1:dim(x)[2],1])
+                                               })))
+
+  sens <-
+    data.frame(
+      varNames = varnames,
+      mean = base::colMeans(sensitivities, na.rm = TRUE),
+      std = sapply(sensitivities,stats::sd, na.rm = TRUE),
+      meanSensSQ = base::colMeans(sensitivities ^ 2, na.rm = TRUE)
+    )
+
+  if (plot) {
+    plotlist <- list()
+
+    plotlist[[1]] <- ggplot2::ggplot(sens) +
+      ggplot2::geom_point(ggplot2::aes_string(x = "mean", y = "std")) +
+      ggplot2::geom_label(ggplot2::aes_string(x = "mean", y = "std", label = "varnames"),
+                          position = "nudge") +
+      ggplot2::geom_point(ggplot2::aes(x = 0, y = 0), size = 5, color = "blue") +
+      ggplot2::geom_hline(ggplot2::aes(yintercept = 0), color = "blue") +
+      ggplot2::geom_vline(ggplot2::aes(xintercept = 0), color = "blue") +
+      # coord_cartesian(xlim = c(min(sens$mean,0)-0.1*abs(min(sens$mean,0)), max(sens$mean)+0.1*abs(max(sens$mean))), ylim = c(0, max(sens$std)*1.1))+
+      ggplot2::labs(x = "mean(Sens)", y = "std(Sens)")
+
+
+    plotlist[[2]] <- ggplot2::ggplot() +
+      ggplot2::geom_col(ggplot2::aes(x = varnames, y = colMeans(sensitivities[, , 1] ^ 2, na.rm = TRUE),
+                                     fill = colMeans(sensitivities[, , 1] ^ 2, na.rm = TRUE))) +
+      ggplot2::labs(x = "Input variables", y = "mean(Sens^2)") + ggplot2::guides(fill = "none")
+
+    der2 <- as.data.frame(sensitivities[, , 1])
+    colnames(der2) <- varnames
+    dataplot <- reshape2::melt(der2, measure.vars = varnames)
+    # bwidth <- sd(dataplot$value)/(1.34*(dim(dataplot)[1]/length(varnames)))
+    # In case the data std is too narrow and erase the data
+    if (any(abs(dataplot$value) > 2*max(sens$std, na.rm = TRUE)) ||
+        max(abs(dataplot$value)) < max(sens$std, na.rm = TRUE)) {
+      plotlist[[3]] <- ggplot2::ggplot(dataplot) +
+        ggplot2::geom_density(ggplot2::aes_string(x = "value", fill = "variable"),
+                              alpha = 0.4,
+                              bw = "bcv") +
+        ggplot2::labs(x = "Sens", y = "density(Sens)") +
+        ggplot2::xlim(-1 * max(abs(dataplot$value), na.rm = TRUE),
+                      1 * max(abs(dataplot$value), na.rm = TRUE))
+    } else {
+      plotlist[[3]] <- ggplot2::ggplot(dataplot) +
+        ggplot2::geom_density(ggplot2::aes_string(x = "value", fill = "variable"),
+                              alpha = 0.4,
+                              bw = "bcv") +
+        ggplot2::labs(x = "Sens", y = "density(Sens)") +
+        ggplot2::xlim(-2 * max(sens$std, na.rm = TRUE), 2 * max(sens$std, na.rm = TRUE))
+    }
+    # Plot the list of plots created before
+    gridExtra::grid.arrange(grobs = plotlist,
+                            nrow  = length(plotlist),
+                            ncols = 1)
+  }
+
+  if (.returnSens) {
+    if(!.rawSens) {
+      # Check if there are more than one output
+      return(sens)
+    } else {
+      # Return sensitivities without processing
+      colnames(sensitivities) <- finalModel$coefnames
+      return(sensitivities)
+    }
+  }
+}
